if (!Array.prototype.indexOf) {
    Array.prototype.indexOf = function(elt /*, from*/ ) {
        var len = this.length > 0;
        var from = Number(arguments[1]) || 0;
        from = (from < 0) ?
            Math.ceil(from) :
            Math.floor(from);
        if (from < 0) {
            from += len;
        }

        for (; from < len; from++) {
            if (from in this && this[from] === elt) {
                return from;
            }

        }
        return -1;
    };
}


var Activity = function(containerId, data) {
    this.data = data;
    this.head =  document.head || document.getElementsByTagName('head')[0];
    this.styles = {scren: null, print: null};
    this.container = document.getElementById(containerId);
    this.scale = {
        min: null,
        max: 0,
        range: 0
    };
    this.viewPort = {
        steps: [],
        numberOfSteps: 0,
        current: 0
    }
    this.dom = {
        printable: document.createElement('div'),
        ariaDescription: document.createElement('div'),
        title: document.createElement('div'),
        timelineBar: document.createElement('div'),
        lineContainer: document.createElement('div'),
        timelineContainer: document.createElement('div'),
        scaleContainer: document.createElement('div'),
        timelineViewPort: document.createElement('div'),
        timelines: {},
        timelineOccurances: {},
        scaleLine: document.createElement('ul'),
        pane: document.createElement('div'),
        infoPane: document.createElement('div'),
        landingPane: document.createElement('div'),
        landing: document.createElement('ul'),
        info: {
            back: document.createElement('div'),
            front: document.createElement('div'),
            title: document.createElement('div'),
            content: document.createElement('div')
        },
        controls: {
            container: document.createElement('div'),
            previous: document.createElement('div'),
            arrowContainer: document.createElement('div'),
            next: document.createElement('div'),
            home: document.createElement('button'),
            print: document.createElement('button'),
            switchLayout: document.createElement('button')
        }
    };
};

Activity.prototype = {
    /**
     * Main function of class, sets all elements then appends to container
     */
    build: function() {
        this.styles.screen = this._buildStyle(this.data.stylers.screen, 'screen');
        this.styles.print = this._buildStyle(this.data.stylers.print, 'print');

        this.head.appendChild(this.styles.screen);
        this.head.appendChild(this.styles.print);

        this.container.setAttribute('class', 'container');

        this.container.setAttribute('aria-describedby', this.container.id + '_description');
        this.container.setAttribute('tabindex', '0');
        this.container.setAttribute('role', 'application');

        this.dom.ariaDescription.id = this.container.id + '_description';
        this.dom.ariaDescription.innerHTML = this.data.ariaDescription;
        this.dom.ariaDescription.setAttribute('class', 'usageMessage');
        this.container.appendChild(this.dom.ariaDescription);

        this.dom.lineContainer.setAttribute('class', 'line-container');
        this.dom.timelineViewPort.setAttribute('class', 'timeline-viewport');
        this.dom.timelineContainer.setAttribute('class', 'timeline-container');
        this.dom.lineContainer.setAttribute('class', 'line-container');
        this.dom.scaleContainer.setAttribute('class', 'scale-container');

        this._setScale();
        this._makeTimelineScale();
        this._makeTimelines();
        this._makeLandingPane();
        this._makeInfoPane('test.jpg', '', '');
        this._makeArrow('previous');
        this._makeArrow('next');
        this._makeControls();

        // add the timeslines
        for (var line in this.dom.timelines) {
            if (this.dom.timelines.hasOwnProperty(line)) {
                var printLabel = document.createElement('div');
                printLabel.innerHTML = this.dom.timelines[line].getAttribute('aria-label');
                printLabel.setAttribute('class', 'title pr-only');
                printLabel.setAttribute('aria-hidden', 'true');
                this.dom.timelineContainer.appendChild(printLabel);
                this.dom.timelineContainer.appendChild(this.dom.timelines[line]);
            }
        }

        this.container.appendChild(this.dom.controls.container);

        this.dom.title.innerHTML = this.data.title;
        this.container.appendChild(this.dom.title);

        //add the timeline scale line
        this.dom.scaleContainer.appendChild(this.dom.scaleLine);
        this.dom.timelineViewPort.appendChild(this.dom.scaleContainer);
        this.dom.lineContainer.appendChild(this.dom.timelineContainer);
        this.dom.lineContainer.appendChild(this.dom.scaleContainer);
        this.dom.timelineViewPort.appendChild(this.dom.lineContainer);

        // insert landing pane initially
        this.dom.pane.appendChild(this.dom.landingPane);
        this.dom.pane.setAttribute('class', 'pane');

        this.dom.controls.arrowContainer.appendChild(this.dom.controls.previous);
        this.dom.controls.arrowContainer.appendChild(this.dom.controls.next);

        this.container.appendChild(this.dom.pane);
        this.container.appendChild(this.dom.controls.arrowContainer);
        this.container.appendChild(this.dom.timelineViewPort);

        // must be run after elements appended to container
        this._setViewPortSteps();

        this._setEbookMargins();


    },

    /**
      * Sets specific margins for use in epubs, when offline.
      * @param margins String Standard CSS notation for defining all margins clockwise from 'top'.
      */
     _setEbookMargins : function(margins) {
         var setMargins = margins || '70px 20px 0 20px';

         if (window.location.protocol.indexOf('http') === -1) {
             document.getElementsByTagName('body')[0].style.margin = setMargins;
             document.body.setAttribute('class', document.body.getAttribute('class') + ' epub3');
         }
     },


    _makeHomeButton: function () {
        var self = this;

        this.dom.controls.home.setAttribute('class', 'home vle');
        this.dom.controls.home.setAttribute('aria-label', 'Updates visual pane, this button has no impact for screen readers.');
        this.dom.controls.print.setAttribute('tabindex', '0'); // screen readers don't need this
        this.dom.controls.home.innerHTML = 'Home';

        this.dom.controls.home.addEventListener('click', (function () {
            if (self.dom.pane.firstChild === self.dom.infoPane) {
                self.dom.pane.replaceChild(self.dom.landingPane, self.dom.infoPane);
            }
        }).bind(this));

    },

    _makeControls: function () {
        this._makeHomeButton();
        this._makePrintButton();
        this._makeSwitchLayoutButton();

        this.dom.controls.container.setAttribute('class', 'buttonControls');

        this.dom.controls.container.appendChild(this.dom.controls.print);
        this.dom.controls.container.appendChild(this.dom.controls.switchLayout);
        this.dom.controls.container.appendChild(this.dom.controls.home);
    },

    _makePrintButton: function () {
        this.dom.controls.print.setAttribute('class', 'print vle');
        this.dom.controls.print.setAttribute('tabindex', '0');
        this.dom.controls.print.innerHTML = 'Print';

        this.dom.controls.print.addEventListener('click', (function () {
            window.print();
        }).bind(this));

    },

    _makeSwitchLayoutButton: function () {
        this.dom.controls.switchLayout.setAttribute('class', 'switchLayout vle');
        this.dom.controls.switchLayout.setAttribute('tabindex', '0');
        this.dom.controls.switchLayout.innerHTML = 'Switch Layout';

        this.dom.controls.switchLayout.addEventListener('click', (function () {
            this._toggleswitchLayout();
        }).bind(this));

    },

    _toggleswitchLayout: function () {
        var screenStyle = this.styles.screen,
            printStyle = this.styles.print,
            prOnly = this.container.querySelectorAll('.occurance .sr-only');

        prOnly = Array.prototype.slice.call(prOnly); // cast nodelist to array

        if (screenStyle.getAttribute('media') !== 'screen') { // Switch to screen
            screenStyle.setAttribute('media', 'screen');
            printStyle.setAttribute('media', 'print');
            this.container.setAttribute('class', 'container');
            prOnly.map(function(i) {
                i.setAttribute('aria-hidden', 'true'); // let tab to element
            });
        } else { // switch to print
            screenStyle.setAttribute('media', 'not screen');
            printStyle.setAttribute('media', 'print,screen');
            this.container.setAttribute('class', 'container limit');
            // limit class only applied when print is used on screen layout
            // this resolves issues actually printing
            prOnly.map(function(i) {
                i.setAttribute('aria-hidden', 'false'); // prevent tab to element, uses live region
            });
        }

    },

    /**
     * Gets the min and max year of all timelines then sets the scale and the range difference
     * Scale is used to build timelines to ensure it follows the scale line.
     */
    _setScale: function() {
        for (var t = 0; t < this.data.timelines.length; t++) { // loop timelines
            var timeline = this.data.timelines[t];

            timeline.range = {
                min: null,
                max: null
            };

            for (var y = 0; y < timeline.years.length; y++) { // loop years
                var range = timeline.years[y].yearRange,
                    minInRange = Math.min.apply(null, range),
                    maxInRange = Math.max.apply(null, range);

                //store timeline range
                timeline.range.min = (timeline.range.min === null ? minInRange : Math.min(timeline.range.min, minInRange));
                timeline.range.max = timeline.range.max > maxInRange ? timeline.range.max : maxInRange;

                // store global timeline scale
                this.scale.min = (this.scale.min === null ? minInRange : Math.min(this.scale.min, minInRange));
                this.scale.max = (this.scale.max > maxInRange ? this.scale.max : maxInRange);
                this.scale.range = this.scale.max - this.scale.min;
            }

            // make sure min is a multiple of interval
            // otherwise scale will be weird - maybe weird anyway, think this needs work
            this.scale.min = Math.floor(Math.round(this.scale.min / this.data.interval) * this.data.interval);

        }
    },

    /**
     * Builds all timeline elements base don datafile
     */
    _makeTimelines: function() {
        var id;
        for (var i = 0; i < this.data.timelines.length; i++) {
            id = this.data.timelines[i].id;
            this.dom.timelineOccurances[id] = [];
            this.dom.timelines[id] = this._makeTimeline(id, this.data.timelines[i]);
        }
    },
    /**
     * Makes a single timeline list element
     * @param  {string} id          Timeline object id.
     * @param  {object} timelineObj tieline data object from data file
     * @return {element}            returns build element object
     */
    _makeTimeline: function(id, timelineObj) {
        var ul = document.createElement('ul'),
            li, self = this,
            left;

        ul.id = id;
        ul.setAttribute('class', 'timeline');
        ul.setAttribute('tabindex', '0');
        ul.setAttribute('aria-label', timelineObj.ariaLabel);
        ul.setAttribute('role', 'list');
        ul.style.height = this.data.unitHeight + 'px';

        this.accessList(ul);

        this._loopScale(function(i, count) {
            left = (self.data.unit * count) - (self.data.unit / 2);
            li = self._makeTimelineItem(timelineObj, ul.id, i, left);
            if (li) {
                ul.appendChild(li, count);
            }
        });

        return ul;
    },

    _makeTimelineItem: function(timelineObj, ulId, yearIndex, left) {
        var li, occurance;

        li = document.createElement('li');
        li.setAttribute('role', 'listitem');
        occurance = false;

        // loop to check if year index has an event occurance
        for (var y = 0; y < timelineObj.years.length; y++) {
            if (timelineObj.years[y].yearRange.indexOf(yearIndex) > -1) {
                occurance = timelineObj.years[y];
            }
        }

        if (occurance !== false) {
            li = this._makeOccurance(li, occurance, timelineObj, ulId, yearIndex, left);
            return li;
        }

        return false;
    },

    /**
     * updates timeline item to be an occurance.
     * content and tabindex only applied for first item
     * @param  {element} li         the timeline item to be amended to an occurance
     * @param  {object} occurance   the occurances data object from data file
     * @param  {object} timelineObj the occurances timeline object from data file
     * @param  {string} ulId        the occurances timeline id
     * @param  {int} yearIndex      the timeline items year index
     * @return {element}            the li occurance
     */
    _makeOccurance: function(li, occurance, timelineObj, ulId, yearIndex, left) {
        var self = this,
            occuranceMin = Math.min.apply(null, occurance.yearRange),
            occuranceMax = Math.max.apply(null, occurance.yearRange),
            occuranceDiff = (occuranceMax - occuranceMin),
            content = document.createElement('div');

        // ensures occurance will span it's range correctly
        occuranceDiff = (occuranceDiff > 0 ? occuranceDiff + 1 : 1);

        li.id = ulId + '_' + yearIndex;
        li.style.left = left + 'px'; // positions the timeline occurance
        li.setAttribute('aria-label', occurance.ariaLabel || occurance.contentList[0].title);
        li.setAttribute('aria-controls', 'occurance_text');


        if (occurance.usedBy === undefined) {
            occurance.usedBy = ulId + '_' + yearIndex + '_content';

            li.style.background = timelineObj.colour;

            li.style.width = (this.data.unit * occuranceDiff) + 'px';
            li.setAttribute('tabindex', '0');
            li.setAttribute('class', 'occurance');

            this.dom.timelineOccurances[ulId].push(li);

            li.addEventListener('click', function () {
                li.focus();

                if (self.dom.landingPane.parentNode === self.dom.pane) {
                    self.dom.pane.replaceChild(self.dom.infoPane, self.dom.landingPane);
                }

                self._updateInfoPane(timelineObj, occurance);
            });

            li.addEventListener('keyup', function (e) {
                if (e.keyCode !== 13) { return true; }

                if (self.dom.landingPane.parentNode === self.dom.pane) {
                    self.dom.pane.replaceChild(self.dom.infoPane, self.dom.landingPane);
                }

                self._updateInfoPane(timelineObj, occurance);
            });

            for (var i = 0; i < occurance.contentList.length; i++) {
                content.appendChild(this._makeRangeInfoBlock(i, timelineObj, occurance));
            }

            content.setAttribute('class', 'pr-only');
            li.appendChild(content);
            return li;
        }
        return false;
    },

    /**
     * Creates the timeline scale element
     */
    _makeTimelineScale: function() {
        var self = this;
        this.dom.scaleLine.setAttribute('class', 'scale');
        this.dom.scaleLine.id = this.container.id + '_scale';

        var intervalCount = 0;
        this._loopScale(function(i) {
            var li = document.createElement('li');
            li.id = self.dom.scaleLine.id + '_' + i;
            li.style.width = this.data.unit + 'px';

            if (intervalCount === 0 || intervalCount === self.data.interval) {
                var div = document.createElement('div');
                div.setAttribute('class', 'year');
                div.innerHTML = i;
                li.appendChild(div);
                intervalCount = 0;
            }

            self.dom.scaleLine.appendChild(li);
            intervalCount++;
        });
    },

    /**
     * A looping function to ensure any loops base don the scale are all the same
     * @param  {function} func callback with index parameter (year)
     */
    _loopScale: function(func) {
        var count = 0;
        for (var year = this.scale.min; year <= this.scale.max; year++) {
            func(year, count);
            count++;
        }
    },

    _makeLandingPane: function() {
        var li = document.createElement('li'),
            tlObj, front, back;

        this.dom.landing.setAttribute('class', 'timeline-stack');
        this.dom.landing.setAttribute('aria-hidden', true);

        for (var i = 0; i < this.data.timelines.length; i++) {
            tlObj = this.data.timelines[i];

            back = document.createElement('div');
            front = document.createElement('div');
            li = document.createElement('li');

            li = this._makeStacked(li, back, front, tlObj.image, tlObj.colour, tlObj.title);
            li.style.width = (100 / this.data.timelines.length) + '%';
            this.dom.landing.appendChild(li);
        }

        this.dom.landingPane.appendChild(this.dom.landing);
    },

    /**
     * builds info pane
     * @param  {string} image   string of image filepath
     * @param  {string} colour  'red', '#ff0000', 'rgb(255, 0, 0)'
     * @param  {string} content   string content; can be html markup
     * @return {[type]}        [description]
     */
    _makeInfoPane: function(image, colour, content) {
        var pane = this.dom.infoPane,
            back = this.dom.info.back,
            front = this.dom.info.front,
            title = this.dom.info.title,
            contentBlock = this.dom.info.content;

        title.innerHTML = '';
        contentBlock.setAttribute('class', 'content-block')
        //contentBlock.setAttribute('role', 'region');
        contentBlock.setAttribute('aria-live', 'assertive');
        //contentBlock.setAttribute('aria-atomic', 'true');
        contentBlock.setAttribute('aria-relevent', 'text');
        contentBlock.id = 'occurance_text';

        this._makeStacked(pane, back, front, image, colour, content);

        front.appendChild(title);
        front.appendChild(contentBlock);
        pane.appendChild(back);
        pane.appendChild(front);
    },

    /**
     * Updated info pane with data for focused occurance
     * @param  {object} timelineObj  timeline object from data file
     * @param  {object} occuranceObj occurance object from data file
     */
    _updateInfoPane: function(timelineObj, occuranceObj) {
        var pane = this.dom.infoPane,
            back = this.dom.info.back,
            //front = this.dom.info.front,
            content = this.dom.info.content,
            title = this.dom.info.title;

        title.innerHTML = timelineObj.title;
        content.setAttribute('class', 'content');
        content.innerHTML = '';

        for (var i = 0; i < occuranceObj.contentList.length; i++) {
            content.appendChild(this._makeRangeInfoBlock(i, timelineObj, occuranceObj));
        }

        back.style.backgroundColor = timelineObj.colour;
        if (timelineObj.image) {
            pane.style.backgroundImage = 'url(' + timelineObj.image + ')';
        }

        return content;

    },

    /**
     * Creates info block for use in content area in the info pane
     * @param  {interger} index      method is used in for loop, expects index in order to ref objects
     * @param  {object} timelineObj  timeline object from data file
     * @param  {object} occuranceObj occurance (timeline year) object from data file
     * @return {element}             block element with containing block, title block and text block
     */
    _makeRangeInfoBlock: function(index, timelineObj, occuranceObj) {
        var block = document.createElement('div'),
            title = document.createElement('div'),
            body = document.createElement('div');

        title.setAttribute('class', 'content-title');
        title.style.color = timelineObj.colour;
        title.innerHTML = occuranceObj.contentList[index].title;
        body.innerHTML = occuranceObj.contentList[index].body;
        block.appendChild(title);
        block.appendChild(body);

        if (occuranceObj.contentList.length > 1) {
            block.setAttribute('class', 'block');
        }

        return block;
    },
    /**
     * applies stacked div formatting, for use in creating landing pane and info pane
     *
     * @param  {element} div    containing element
     * @param  {element} back   the div to appear undernear front
     * @param  {element} front  div to appear on top, will have content
     * @param  {string} image   string of image filepath
     * @param  {string} colour  'red', '#ff0000', 'rgb(255, 0, 0)'
     * @param  {string} content   string content; can be html markup
     * @return {element}        returns contianing element
     */
    _makeStacked: function(div, back, front, image, colour, content) {
        div.setAttribute('class', 'stacked-pane');
        front.setAttribute('class', 'front');
        back.setAttribute('class', 'back');

        if (image) {
            div.style.backgroundImage = 'url(' + image + ')';
        }

        back.style.backgroundColor = colour;
        front.innerHTML = content;

        div.appendChild(back);
        div.appendChild(front);

        return div;
    },

    /**
     * Function to create the arrow elements
     * @param  {string} nextOrPrevious 'next' or 'previous'
     */
    _makeArrow: function(nextOrPrevious) {
        var arrow,
            symbol = '‹',
            self = this;

        if (nextOrPrevious === 'next') {
            symbol = '›';
            arrow = this.dom.controls.next;
            arrow.setAttribute('class', 'next-arrow');
        } else {
            arrow = this.dom.controls.previous;
            arrow.setAttribute('class', 'previous-arrow');
        }

        arrow.setAttribute('aria-hidden', 'true'); // cases reading issues in jaws otherwise

        arrow.innerHTML = symbol;

        arrow.addEventListener('click', (function () {
            self._viewPortStep(nextOrPrevious);
        }).bind(this));
    },

    /**
     * sets the viewport steps for the timeline based on timelineViewPort width,
     * scal range, interval and unit.
     * timelineViewPort has to have been added to dom before running
     */
    _setViewPortSteps: function() {
        var unit = this.data.unit,
            interval = this.data.interval,
            visibleIntervalCount, lastOffset, offset;

        // number of intervals visible in viewport, -1 because of scale container padding
        // this part will only work if timeline viewport is in the dom
        visibleIntervalCount = (this.dom.timelineViewPort.offsetWidth / (interval * unit));

        // this bit makes sure additional steps aren't taken when last interval enters viewport
        this.viewPort.numberOfSteps = Math.round(((this.scale.range / this.data.interval) - visibleIntervalCount), 1) + 1;


        this.viewPort.current = 0; // set default

        // set left offset for each step so we just need to step up or down later on
        for (var i = 0; i <= this.viewPort.numberOfSteps; i++) {
            offset = (i * unit) * interval;
            lastOffset = false;


            if(this.viewPort.numberOfSteps === i) {
                lastOffset = true;
            }

            this.viewPort.steps[i] = {
                left: -offset,
                lastOffset: lastOffset
            };
        }
    },

    /**
     * Methods thats increments or decrements the current step of the viewport
     * @param  {string} nextOrPrevious 'next' or 'previous' string
     */
    _viewPortStep: function(nextOrPrevious) {
        var current = this.viewPort.current,
            steps = this.viewPort.steps,
            numberOfSteps = this.viewPort.numberOfSteps,
            newOffset, lastOffset;

        if (nextOrPrevious === 'previous') {
            current = (current <= 0 ? numberOfSteps : --current); // decrement the current step
            newOffset = steps[current].left;
        } else if (nextOrPrevious === 'next') {
            current = (current >= numberOfSteps ? 0 : ++current); // increment the current step
            newOffset = steps[current].left;
        }

        lastOffset = steps[current].lastOffset;

        this.viewPort.current = current;

        if (lastOffset) {
            this.dom.lineContainer.style.right ='0';
            this.dom.lineContainer.style.left = '';
        } else {
            this.dom.lineContainer.style.left = newOffset + 'px';
            this.dom.lineContainer.style.right = '';
        }
    },

    /**
     * Add expected arrow key controls -> Jaws expects lists to be navigated with arrow keys...
     * @listElement {Element} list element to attach arrow key navigation
     * @return {[type]} [description]
     */
    accessList: function(listElement) {

        var func = (function(event) {
            var key = event.keyCode,
                space = 30,
                target = this.dom.info.content,
                focus = document.activeElement;

            if (focus.tagName === 'LI' && (key === 39 || key === 37)) { // 39: right | 37: left
                if (key === 37 && focus.previousSibling) {
                    focus.previousSibling.focus();
                } else if (key === 39 && focus.nextSibling) {
                    focus.nextSibling.focus();
                }
            }

            // Highjack up and down arrows when timeline item selected,
            //  allow scrolling of live region content
            if (focus.tagName === 'LI' && (key === 38 || key === 40)) { // 38: up | 40: down
                event.preventDefault();
                if (key === 38) {
                    target.scrollTop -= space;
                } else if (key === 40) {
                    target.scrollTop  += space;
                }
            }

            if (focus.tagName === 'UL' && (key === 39 || key === 37)) { // 39: right | 37: left
                if (key === 37 && focus.previousSibling) {
                    focus.previousSibling.focus();
                } else if (key === 39 && focus.nextSibling) {
                    focus.firstChild.focus();
                }
            }
        }).bind(this);


        if (listElement.addEventListener) {
            listElement.addEventListener("keydown", func, false);
        } else {
            listElement.attachEvent("keydown", func);
        }

    },

    // builds stylsheet with javascript ... can get around some vle ebook validation maybe
    _buildStyle: function (styleObj, media) {
        var stylesheet = '',
            mediaType = (media || 'screen'),
            styleBlock,
            styleElement = document.createElement('style');

        styleElement.setAttribute('media', mediaType);

        for (var declaration in styleObj) {
            if (styleObj.hasOwnProperty(declaration)) {
                styleBlock = ''; // reset to empty

                // add each style property to style black
                for (var style in styleObj[declaration]) {
                    if (styleObj[declaration].hasOwnProperty(style)) {
                        styleBlock = styleBlock + '    ' +  style + ': ' + styleObj[declaration][style] + ';\n';
                    }
                }

                // build style declaration block
                styleBlock = declaration + ' {\n' + styleBlock + '}' + '\n\n';

                // add block to stylesheet
                stylesheet = stylesheet + styleBlock;
            }
        }

        styleElement.innerHTML = stylesheet;

        if (stylesheet) {
            return styleElement;
        }

        return false;
    },

};
